@@ -101,10 +101,15 @@
  * Last reviewed on 2005-11-23 (0.9.5)
  */
 
+#include "gst_private.h"
 #include <stdlib.h>
 #include <stdio.h>
+#include <sys/types.h>
+#ifdef HAVE_FORK
+#include <sys/wait.h>
+#endif //HAVE_FORK
+#include <unistd.h>
 
-#include "gst_private.h"
 #include "gst-i18n-lib.h"
 #include <locale.h>             /* for LC_ALL */
 
@@ -117,6 +122,10 @@
 
 static gboolean gst_initialized = FALSE;
 
+#ifndef GST_DISABLE_REGISTRY
+static GList *plugin_paths = NULL;      /* for delayed processing in post_init */
+#endif
+
 extern gint _gst_trace_on;
 
 /* set to TRUE when segfaults need to be left as is */
@@ -242,10 +251,6 @@ parse_debug_list (const gchar * list)
 }
 #endif
 
-#ifndef GST_HAVE_GLIB_2_8
-#define G_OPTION_FLAG_NO_ARG 0
-#endif
-
 /**
  * gst_init_get_option_group:
  *
@@ -394,8 +399,8 @@ gst_init (int *argc, char **argv[])
 static void
 add_path_func (gpointer data, gpointer user_data)
 {
-  GST_INFO ("Adding plugin path: \"%s\"", (gchar *) data);
-  gst_registry_scan_path (gst_registry_get_default (), (gchar *) data);
+  GST_INFO ("Adding plugin path: \"%s\", will scan later", (gchar *) data);
+  plugin_paths = g_list_append (plugin_paths, g_strdup (data));
 }
 #endif
 
@@ -463,17 +468,10 @@ split_and_iterate (const gchar * stringlist, gchar * separator, GFunc iterator,
 static gboolean
 init_pre (void)
 {
-#ifdef GST_HAVE_GLIB_2_8
   /* GStreamer was built against a GLib >= 2.8 and is therefore not doing
    * the refcount hack. Check that it isn't being run against an older GLib */
   if (glib_major_version < 2 ||
       (glib_major_version == 2 && glib_minor_version < 8)) {
-#else
-  /* GStreamer was built against a GLib < 2.8 and is therefore doing
-   * the refcount hack. Check that it isn't being run against a newer GLib */
-  if (glib_major_version > 2 ||
-      (glib_major_version == 2 && glib_minor_version >= 8)) {
-#endif
     g_warning ("GStreamer was compiled against GLib %d.%d.%d but is running"
         " against %d.%d.%d. This will cause reference counting issues",
         GLIB_MAJOR_VERSION, GLIB_MINOR_VERSION, GLIB_MICRO_VERSION,
@@ -602,6 +600,21 @@ init_post (void)
     char *registry_file;
     const char *plugin_path;
     GstRegistry *default_registry;
+    gboolean changed = FALSE;
+    GList *l;
+
+    for (l = plugin_paths; l != NULL; l = l->next) {
+      GST_INFO ("Scanning plugin path: \"%s\"", (gchar *) l->data);
+      /* CHECKME: add changed |= here as well? */
+      gst_registry_scan_path (gst_registry_get_default (), (gchar *) l->data);
+      g_free (l->data);
+    }
+    g_list_free (plugin_paths);
+    plugin_paths = NULL;
+
+#ifdef HAVE_FORK
+    pid_t pid;
+#endif
 
     default_registry = gst_registry_get_default ();
     registry_file = g_strdup (g_getenv ("GST_REGISTRY"));
@@ -609,59 +622,104 @@ init_post (void)
       registry_file = g_build_filename (g_get_home_dir (),
           ".gstreamer-" GST_MAJORMINOR, "registry." HOST_CPU ".xml", NULL);
     }
-    GST_DEBUG ("Reading registry cache");
-    gst_registry_xml_read_cache (default_registry, registry_file);
+#ifdef HAVE_FORK
+    /* We fork here, and let the child read and possibly rebuild the registry.
+     * After that, the parent will re-read the freshly generated registry. */
 
-    /* GST_PLUGIN_PATH specifies a list of directories to scan for
-     * additional plugins.  These take precedence over the system plugins */
-    plugin_path = g_getenv ("GST_PLUGIN_PATH");
-    if (plugin_path) {
-      char **list;
-      int i;
-
-      GST_DEBUG ("GST_PLUGIN_PATH set to %s", plugin_path);
-      list = g_strsplit (plugin_path, G_SEARCHPATH_SEPARATOR_S, 0);
-      for (i = 0; list[i]; i++) {
-        gst_registry_scan_path (default_registry, list[i]);
-      }
-      g_strfreev (list);
-    } else {
-      GST_DEBUG ("GST_PLUGIN_PATH not set");
+    GST_DEBUG ("forking");
+    pid = fork ();
+    if (pid == -1) {
+      GST_ERROR ("Failed to fork()");
+      return FALSE;
     }
 
-    /* GST_PLUGIN_SYSTEM_PATH specifies a list of plugins that are always
-     * loaded by default.  If not set, this defaults to the system-installed
-     * path, and the plugins installed in the user's home directory */
-    plugin_path = g_getenv ("GST_PLUGIN_SYSTEM_PATH");
-    if (plugin_path == NULL) {
-      char *home_plugins;
+    if (pid == 0) {
+      /* this is the child */
+#endif /* HAVE_FORK */
 
-      GST_DEBUG ("GST_PLUGIN_SYSTEM_PATH not set");
+      GST_DEBUG ("child reading registry cache");
+      gst_registry_xml_read_cache (default_registry, registry_file);
 
-      /* plugins in the user's home directory take precedence over
-       * system-installed ones */
-      home_plugins = g_build_filename (g_get_home_dir (),
-          ".gstreamer-" GST_MAJORMINOR, "plugins", NULL);
-      gst_registry_scan_path (default_registry, home_plugins);
-      g_free (home_plugins);
+      /* GST_PLUGIN_PATH specifies a list of directories to scan for
+       * additional plugins.  These take precedence over the system plugins */
+      plugin_path = g_getenv ("GST_PLUGIN_PATH");
+      if (plugin_path) {
+        char **list;
+        int i;
 
-      /* add the main (installed) library path */
-      gst_registry_scan_path (default_registry, PLUGINDIR);
-    } else {
-      char **list;
-      int i;
-
-      GST_DEBUG ("GST_PLUGIN_SYSTEM_PATH set to %s", plugin_path);
-      list = g_strsplit (plugin_path, G_SEARCHPATH_SEPARATOR_S, 0);
-      for (i = 0; list[i]; i++) {
-        gst_registry_scan_path (default_registry, list[i]);
+        GST_DEBUG ("GST_PLUGIN_PATH set to %s", plugin_path);
+        list = g_strsplit (plugin_path, G_SEARCHPATH_SEPARATOR_S, 0);
+        for (i = 0; list[i]; i++) {
+          changed |= gst_registry_scan_path (default_registry, list[i]);
+        }
+        g_strfreev (list);
+      } else {
+        GST_DEBUG ("GST_PLUGIN_PATH not set");
       }
-      g_strfreev (list);
+
+      /* GST_PLUGIN_SYSTEM_PATH specifies a list of plugins that are always
+       * loaded by default.  If not set, this defaults to the system-installed
+       * path, and the plugins installed in the user's home directory */
+      plugin_path = g_getenv ("GST_PLUGIN_SYSTEM_PATH");
+      if (plugin_path == NULL) {
+        char *home_plugins;
+
+        GST_DEBUG ("GST_PLUGIN_SYSTEM_PATH not set");
+
+        /* plugins in the user's home directory take precedence over
+         * system-installed ones */
+        home_plugins = g_build_filename (g_get_home_dir (),
+            ".gstreamer-" GST_MAJORMINOR, "plugins", NULL);
+        changed |= gst_registry_scan_path (default_registry, home_plugins);
+        g_free (home_plugins);
+
+        /* add the main (installed) library path */
+        changed |= gst_registry_scan_path (default_registry, PLUGINDIR);
+      } else {
+        char **list;
+        int i;
+
+        GST_DEBUG ("GST_PLUGIN_SYSTEM_PATH set to %s", plugin_path);
+        list = g_strsplit (plugin_path, G_SEARCHPATH_SEPARATOR_S, 0);
+        for (i = 0; list[i]; i++) {
+          changed |= gst_registry_scan_path (default_registry, list[i]);
+        }
+        g_strfreev (list);
+      }
+
+      if (changed) {
+        GST_DEBUG ("writing registry cache");
+        gst_registry_xml_write_cache (default_registry, registry_file);
+      }
+
+      _gst_registry_remove_cache_plugins (default_registry);
+
+#ifdef HAVE_FORK
+      exit (0);
+    } else {
+      /* parent */
+      int status;
+      pid_t ret;
+
+      GST_DEBUG ("parent waiting on child");
+      ret = waitpid (pid, &status, 0);
+      GST_DEBUG ("parent done waiting on child");
+      if (ret == -1) {
+        GST_ERROR ("error during waitpid: %s", g_strerror (errno));
+        return FALSE;
+      }
+
+      if (!WIFEXITED (status)) {
+        GST_ERROR ("child did not exit normally, status: %d", status);
+        return FALSE;
+      }
+
+      GST_DEBUG ("child exited normally with return value %d",
+          WEXITSTATUS (status));
+      GST_DEBUG ("parent reading registry cache");
+      gst_registry_xml_read_cache (default_registry, registry_file);
     }
-
-    gst_registry_xml_write_cache (default_registry, registry_file);
-
-    _gst_registry_remove_cache_plugins (default_registry);
+#endif /* HAVE_FORK */
 
     g_free (registry_file);
   }
